/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const pwt = require('./lib/index');
const { defineConfig } = require('./lib/common/configLoader');
const playwright = require('./index');
const combinedExports = {
  ...playwright,
  ...pwt,
  defineConfig,
};

Object.defineProperty(combinedExports, '__esModule', { value: true });

module.exports = combinedExports;
