/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { test, expect } from './pageTest';

test('should count objects', async ({ page, browserName }) => {
  test.skip(browserName !== 'chromium');
  await page.setContent('<button>Submit</button>');
  await page.evaluate(() => document.querySelectorAll('button'));
  const proto = await page.evaluateHandle(() => HTMLButtonElement.prototype);
  expect(await (proto as any)._objectCount()).toBe(1);
});
