<template>
  <header>
    <img alt="Vue logo" class="logo" src="./assets/logo.svg" width="125" height="125" />
    <router-link to="/">Login</router-link>
    <router-link to="/dashboard">Dashboard</router-link>
  </header>
  <router-view />
</template>
