#!/usr/bin/env node
/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const { execSync } = require('child_process');
const path = require('path');

// Note: we ignore stdout for smaller output.
try {
  execSync('npm run test -- --list --forbid-only', {
    stdio: ['ignore', 'ignore', 'inherit'],
    cwd: path.join(__dirname, '..'),
  });
  execSync('npm run ttest -- --list --forbid-only', {
    stdio: ['ignore', 'ignore', 'inherit'],
    cwd: path.join(__dirname, '..'),
  });
} catch (e) {
  process.exit(1);
}
