:root {
  --non-monospace: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
  --monospace: Consolas, Menlo, monospace;
  font-size: 14px;
}

body {
  font-family: var(--non-monospace);
}

.red { background-color: #ffcdd2; }
.green { background-color: #a5d6a7; }
.blue { background-color: #64b5f6; }
