#!/usr/bin/env node
/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const pw = require('playwright-core');

async function browserVersion(browserType) {
  const browser = await browserType.launch();
  const result = browser.version();
  await browser.close();
  return result;
}

(async () => {
  console.log('- Chromium ' + await browserVersion(pw.chromium));
  console.log('- Mozilla Firefox ' + await browserVersion(pw.firefox));
  console.log('- WebKit ' + await browserVersion(pw.webkit));
})();
